@@ -35,19 +35,19 @@
 #include "thread_rt.h"
 
 #include <sys/time.h>
+#include <semaphore.h>
 
 /* Environment setting */
 unsigned int PWQ_RT_THREADS = 0;
-time_t PWQ_SPIN_USEC = 10000; // The number of microseconds we should spin loop if desired
 unsigned int PWQ_SPIN_THREADS = 0; // The number of threads that should be kept spinning
 unsigned volatile int current_threads_spinning = 0; // The number of threads currently spinning
 
 /* Tunable constants */
 
-#define WORKER_IDLE_SECONDS_THRESHOLD 5
+#define WORKER_IDLE_SECONDS_THRESHOLD 15
 
 /* Function prototypes */
-static unsigned int get_load_average(void);
+static unsigned int get_runqueue_length(void);
 static void * worker_main(void *arg);
 static void * overcommit_worker_main(void *arg);
 static unsigned int get_process_limit(void);
@@ -56,6 +56,7 @@ static void manager_start(void);
 static unsigned int      cpu_count;
 static unsigned int      worker_min;
 static unsigned int      worker_idle_threshold; // we don't go down below this if we had to increase # workers
+static unsigned int      pending_thread_create;
 
 /* Overcommit */
 static struct _pthread_workqueue *ocwq[PTHREAD_WORKQUEUE_MAX];
@@ -63,6 +64,7 @@ static int               ocwq_mask;
 static pthread_mutex_t   ocwq_mtx;
 static pthread_cond_t    ocwq_has_work;
 static unsigned int      ocwq_idle_threads;
+static unsigned int      ocwq_signal_count;     // number of times OCWQ condition was signalled
 
 /* Non-overcommit */
 static struct _pthread_workqueue *wqlist[PTHREAD_WORKQUEUE_MAX];
@@ -74,14 +76,24 @@ static int               wqlist_has_manager;
 static pthread_attr_t    detached_attr;
 
 static struct {
-    volatile unsigned int load,
+    volatile unsigned int runqueue_length,
                     count,
                     idle;
-    unsigned int    sb_wake_pending;
-    pthread_mutex_t sb_wake_mtx;
-    pthread_cond_t  sb_wake_cond;
+    sem_t sb_sem;
+    unsigned int sb_suspend;
 } scoreboard;
 
+/* Thread limits */
+#define DEFAULT_PROCESS_LIMIT 100
+
+#if defined(__sun)
+#include <rctl.h>
+
+#define MIN_PROCESS_LIMIT 4
+#define MAX_PROCESS_LIMIT 1000
+#define THREADS_RESERVED  5 // arbitrary extra thread reservation
+#endif
+
 static unsigned int 
 worker_idle_threshold_per_cpu(void)
 {
@@ -93,16 +105,16 @@ worker_idle_threshold_per_cpu(void)
         case 4:
           return 2;
         case 6:
-          return 3;
+          return 4;
         case 8:
         case 12:
-          return 4;
+          return 6;
         case 16:
         case 24:
-          return 6;
+          return 8;
         case 32:
         case 64:
-          return 8;
+          return 12;
         default:
             return cpu_count / 4;
     }
@@ -110,12 +122,14 @@ worker_idle_threshold_per_cpu(void)
     return 2;
 }
 
+#if !defined(__ANDROID__)
 static void
 manager_reinit(void)
 {
     if (manager_init() < 0)
         abort();
 }
+#endif
 
 int
 manager_init(void)
@@ -125,11 +139,13 @@ manager_init(void)
 
     pthread_mutex_init(&wqlist_mtx, NULL);
     wqlist_mask = 0;
-
+    pending_thread_create = 0;
+    
     pthread_cond_init(&ocwq_has_work, NULL);
     pthread_mutex_init(&ocwq_mtx, NULL);
     ocwq_mask = 0;
     ocwq_idle_threads = 0;
+    ocwq_signal_count = 0;
 
     witem_cache_init();
 
@@ -139,21 +155,62 @@ manager_init(void)
     pthread_attr_setdetachstate(&detached_attr, PTHREAD_CREATE_DETACHED);
 
     /* Initialize the scoreboard */
-    pthread_cond_init(&scoreboard.sb_wake_cond, NULL);
-    pthread_mutex_init(&scoreboard.sb_wake_mtx, NULL);
-
+    
+    if (sem_init(&scoreboard.sb_sem, 0, 0) != 0) {
+        dbg_perror("sem_init()");
+        return (-1);
+    }
+    
+    scoreboard.count = 0;
+    scoreboard.idle = 0;
+    scoreboard.sb_suspend = 0;
+    
     /* Determine the initial thread pool constraints */
     worker_min = 2; // we can start with a small amount, worker_idle_threshold will be used as new dynamic low watermark
-    worker_idle_threshold = worker_idle_threshold_per_cpu();
+    worker_idle_threshold = (PWQ_ACTIVE_CPU > 0) ? (PWQ_ACTIVE_CPU) : worker_idle_threshold_per_cpu();
 
+/* FIXME: should test for symbol instead of for Android */
+#if !defined(__ANDROID__)
     if (pthread_atfork(NULL, NULL, manager_reinit) < 0) {
         dbg_perror("pthread_atfork()");
         return (-1);
     }
+#endif
 
     return (0);
 }
 
+/* FIXME: should test for symbol instead of for Android */
+#if defined(__ANDROID__)
+
+#include <fcntl.h>
+
+int getloadavg(double loadavg[], int nelem)
+{
+   int fd;
+   ssize_t len;
+   char buf[80];
+
+   /* FIXME: this restriction allows the code to be simpler */
+   if (nelem != 1)
+       return (-1);
+
+   fd = open("/proc/loadavg", O_RDONLY);
+   if (fd < 0) 
+       return (-1); 
+
+   len = read(fd, &buf, sizeof(buf));
+   (void) close(fd);
+   if (len < 0) 
+       return (-1);
+
+   if (sscanf(buf, "%lf ", &loadavg[0]) < 1)
+       return (-1);
+
+   return (0); 
+}
+#endif /* defined(__ANDROID__) */
+
 void
 manager_workqueue_create(struct _pthread_workqueue *workq)
 {
@@ -165,79 +222,26 @@ manager_workqueue_create(struct _pthread_workqueue *workq)
         if (ocwq[workq->queueprio] == NULL) {
             ocwq[workq->queueprio] = workq;
             workq->wqlist_index = workq->queueprio;
+            dbg_printf("created workqueue (ocommit=1, prio=%d)", workq->queueprio);
         } else {
-            puts("queue already exists\n");
+            printf("oc queue %d already exists\n", workq->queueprio);
             abort();
         }
     } else {
         if (wqlist[workq->queueprio] == NULL) {
             wqlist[workq->queueprio] = workq; //FIXME: sort by priority
             workq->wqlist_index = workq->queueprio;
+            dbg_printf("created workqueue (ocommit=0, prio=%d)", workq->queueprio);
         } else {
-            puts("queue already exists\n");
+            printf("queue %d already exists\n", workq->queueprio);
             abort();
         }
     }
     pthread_mutex_unlock(&wqlist_mtx);
 }
 
-static struct work *
-wqlist_scan(int *queue_priority)
-{
-    pthread_workqueue_t workq;
-    struct work *witem;
-    int idx;
-
-    idx = ffs(wqlist_mask);
-    if (idx == 0)
-        return (NULL);
-
-    workq = wqlist[idx - 1];
- 
-    pthread_spin_lock(&workq->mtx);
-
-    witem = STAILQ_FIRST(&workq->item_listhead);
-    if (witem != NULL) {
-        STAILQ_REMOVE_HEAD(&workq->item_listhead, item_entry);
-        if (STAILQ_EMPTY(&workq->item_listhead))
-        {
-            unsigned int wqlist_index_bit = (0x1 << workq->wqlist_index);
-            unsigned int new_mask;
-            // Remove this now empty wq from the mask, the only contention here is with threads performing the same
-            // operation on another workqueue, so we will not be long
-            // the 'bit' for this queue is protected by the spin lock, so we will only clear a bit which we have 
-            // ownership for (see additem() below for the corresponding part on the producer side)
-            do
-            {
-                new_mask = atomic_and(&wqlist_mask, ~(wqlist_index_bit));
-            } while (new_mask & wqlist_index_bit);
-        }
-        if (queue_priority != NULL)
-            *queue_priority = workq->queueprio;
-
-        pthread_spin_unlock(&workq->mtx);
-        return (witem);
-    } else {
-        // this could happen if multiple threads raced and found the same bit with ffs() and
-        // emptied the queue completely, so we should just bail out
-        pthread_spin_unlock(&workq->mtx);
-        return (NULL);
-    }
-}
-
-static void _wakeup_manager(void)
-{
-    dbg_puts("asking manager to wake up");
-
-    pthread_mutex_lock(&scoreboard.sb_wake_mtx);
-    scoreboard.sb_wake_pending = 1;
-    pthread_cond_signal(&scoreboard.sb_wake_cond);
-    pthread_mutex_unlock(&scoreboard.sb_wake_mtx);
-    return;
-}
-
 static void *
-overcommit_worker_main(void *arg)
+overcommit_worker_main(void *unused __attribute__ ((unused)))
 {
     struct timespec ts;
     pthread_workqueue_t workq;
@@ -245,8 +249,11 @@ overcommit_worker_main(void *arg)
     void *func_arg;
     struct work *witem;
     int rv, idx;
+    sigset_t sigmask;
 
-    (void)arg;
+    /* Block all signals */
+    sigfillset(&sigmask);
+    pthread_sigmask(SIG_BLOCK, &sigmask, NULL);
      
     pthread_mutex_lock(&ocwq_mtx);
 
@@ -278,133 +285,184 @@ overcommit_worker_main(void *arg)
         ocwq_idle_threads++;
         dbg_printf("waiting for work (idle=%d)", ocwq_idle_threads);
         rv = pthread_cond_timedwait(&ocwq_has_work, &ocwq_mtx, &ts);
-        if (rv != 0) {
+        if (ocwq_signal_count > 0) {
+            ocwq_signal_count--;
+            continue;
+        }
+
+        if ((rv == 0) || (rv == ETIMEDOUT)) {
             /* Normally, the signaler will decrement the idle counter,
                but this path is not taken in response to a signaler.
              */
             ocwq_idle_threads--;
-	    pthread_mutex_unlock(&ocwq_mtx);
-
-            if (rv == ETIMEDOUT) {
-                dbg_puts("timeout, no work available");
-                break;
-            } else {
-                dbg_perror("pthread_cond_timedwait");
-                //TODO: some kind of crash mechanism
-                break;
-            }
+            pthread_mutex_unlock(&ocwq_mtx);
+            break;
         }
+
+        dbg_perror("pthread_cond_timedwait");
+        abort();
+        break;
     }
 
     dbg_printf("worker exiting (idle=%d)", ocwq_idle_threads);
     pthread_exit(NULL);
+    /* NOTREACHED */
+    return (NULL);
+}
+
+static inline void reset_queue_mask(unsigned int wqlist_index)
+{
+    unsigned int wqlist_index_bit = (0x1 << wqlist_index);
+    unsigned int new_mask;
+    
+    // Remove this now empty wq from the mask, the only contention here is with threads performing the same
+    // operation on another workqueue, so we will not be long
+    // the 'bit' for this queue is protected by the spin lock, so we will only clear a bit which we have 
+    // ownership for (see below for the corresponding part on the producer side)
+    
+    new_mask = atomic_and(&wqlist_mask, ~(wqlist_index_bit));
+    
+    while (slowpath(new_mask & wqlist_index_bit))
+    {
+        _hardware_pause();
+        new_mask = atomic_and(&wqlist_mask, ~(wqlist_index_bit));
+    }       
+    
+    return;
+}
+
+static struct work *
+wqlist_scan(int *queue_priority, int skip_thread_exit_events)
+{
+    pthread_workqueue_t workq;
+    struct work *witem = NULL;
+    int idx;
+    
+    idx = ffs(wqlist_mask);
+
+    if (idx == 0)
+        return (NULL);
+    
+    workq = wqlist[idx - 1];
+    
+    pthread_spin_lock(&workq->mtx);
+
+    witem = STAILQ_FIRST(&workq->item_listhead);
+
+    if (witem)
+    {
+        if (!(skip_thread_exit_events && (witem->func == NULL)))
+        {
+            STAILQ_REMOVE_HEAD(&workq->item_listhead, item_entry);
+            
+            if (STAILQ_EMPTY(&workq->item_listhead))
+                reset_queue_mask(workq->wqlist_index);
+            
+            *queue_priority = workq->queueprio;
+        }
+        else
+            witem = NULL;
+    }
+
+    pthread_spin_unlock(&workq->mtx);
+    
+    return (witem); // NULL if multiple threads raced for the same queue 
+}
+
+// Optional busy loop for getting the next item for a while if so configured
+// We'll only spin limited number of threads at a time (this is really mostly useful when running
+// in low latency configurations using dedicated processor sets, usually a single spinner makes sense)
+
+static struct work *
+wqlist_scan_spin(int *queue_priority)
+{
+    struct work *witem = NULL;
+    
+    // Start spinning if relevant, otherwise skip and go through
+    // the normal wqlist_scan_wait slowpath by returning the NULL witem.
+    if (atomic_inc_nv(&current_threads_spinning) <= PWQ_SPIN_THREADS)
+    {
+        while ((witem = wqlist_scan(queue_priority, 1)) == NULL)
+            _hardware_pause();
+        
+        /* Force the manager thread to wakeup if we are the last idle one */
+        if (scoreboard.idle == 1)
+            (void) sem_post(&scoreboard.sb_sem);        
+    }
+    
+    atomic_dec(&current_threads_spinning);    
+
+    return witem;
+}
+
+// Normal slowpath for waiting on the condition for new work
+// here we also exit workers when needed
+static struct work *
+wqlist_scan_wait(int *queue_priority)
+{
+    struct work *witem = NULL;
+    
+    pthread_mutex_lock(&wqlist_mtx);
+    
+    while ((witem = wqlist_scan(queue_priority, 0)) == NULL)
+        pthread_cond_wait(&wqlist_has_work, &wqlist_mtx);
+    
+    pthread_mutex_unlock(&wqlist_mtx);
+
+    /* Force the manager thread to wakeup if we are the last idle one */
+    if (scoreboard.idle == 1)
+        (void) sem_post(&scoreboard.sb_sem);        
+
+    // We only process worker exists from the slow path, wqlist_scan only returns them here
+    if (slowpath(witem->func == NULL)) 
+    {
+        dbg_puts("worker exiting..");
+        atomic_dec(&scoreboard.idle);
+        atomic_dec(&scoreboard.count);
+        witem_free(witem);
+        pthread_exit(0);
+    }
+    
+    return witem;
 }
 
 static void *
-worker_main(void *arg)
+worker_main(void *unused __attribute__ ((unused)))
 {
     struct work *witem;
-    void (*func)(void *);
-    void *func_arg;
-    int queue_priority = 0;
-    struct timespec ts_start, ts_now;
-    
-    (void) arg;
-    dbg_puts("worker thread started");
+    int current_thread_priority = WORKQ_DEFAULT_PRIOQUEUE;
+    int queue_priority = WORKQ_DEFAULT_PRIOQUEUE;
 
-    if (PWQ_RT_THREADS)
-        ptwq_set_current_thread_priority(WORKQ_HIGH_PRIOQUEUE); // start at highest priority possible
+    dbg_puts("worker thread started");
+    atomic_dec(&pending_thread_create);
         
     for (;;) {
 
-        witem = wqlist_scan(&queue_priority);
+        witem = wqlist_scan(&queue_priority, 1); 
 
-        // Only take overhead of sleeping and/or spinning if we 
-        // could not get a witem cheaply using the spinlock above
-        if (slowpath(!witem))
+        if (!witem)
         {
-            // Optional busy loop for getting the next item for a while if so configured
-            // We'll only spin limited thread at a time (this is really mostly useful when running
-            // in low latency configurations using dedicated processor sets)
-            if ((PWQ_SPIN_THREADS > 0) && (current_threads_spinning <= PWQ_SPIN_THREADS))
-            {
-                atomic_inc(&current_threads_spinning);
-
-                // If we are racing with another thread, let's skip
-                // spinning and instead go through the slowpath below
-
-                if (current_threads_spinning <= PWQ_SPIN_THREADS)
-                {
-                    clock_gettime(CLOCK_REALTIME, &ts_start);
-                    ts_now.tv_sec = ts_start.tv_sec;
-                    ts_now.tv_nsec = ts_start.tv_nsec;
-                    
-                    // Spin until we get an item or PWQ_SPIN_USEC microseconds passes
-                    while (!witem && (((ts_now.tv_sec - ts_start.tv_sec) * 1000000) + (((ts_now.tv_nsec - ts_start.tv_nsec) / 1000)) <= PWQ_SPIN_USEC))
-                    {
-                        witem = wqlist_scan(&queue_priority);
-                        if (!witem)
-                        {
-                            // Perhaps a hardware pause
-                            // instruction could be used here to keep the pace down, probably not needed though
-                            clock_gettime(CLOCK_REALTIME, &ts_now);
-                        }
-                    }                
-                }
-                
-                atomic_dec(&current_threads_spinning);
-            }
-
-            // No witem from the busy loop, let's wait for wakeup
+            witem = wqlist_scan_spin(&queue_priority);
+            
             if (!witem)
-            {
-                pthread_mutex_lock(&wqlist_mtx);
-
-                /*
-                  TODO: Consider using pthread_cond_timedwait() so that
-                  workers can self-terminate if they are idle too long.
-                  This would also be a failsafe in case there are bugs
-                  with the scoreboard that cause us to "leak" workers.
-                 */
-                while ((witem = wqlist_scan(&queue_priority)) == NULL)
-                    pthread_cond_wait(&wqlist_has_work, &wqlist_mtx);
-
-                pthread_mutex_unlock(&wqlist_mtx);
-            }
+                witem = wqlist_scan_wait(&queue_priority);
         }
+           
+        if (PWQ_RT_THREADS && (current_thread_priority != queue_priority))
+        {
+            current_thread_priority = queue_priority;
+            ptwq_set_current_thread_priority(current_thread_priority);
+        }
+        
+        /* Invoke the callback function */
 
         atomic_dec(&scoreboard.idle);
-
-        if (slowpath(witem->func == NULL)) {
-            dbg_puts("worker exiting..");
-            atomic_dec(&scoreboard.count);
-            witem_free(witem);
-            pthread_exit(0);
-        }
-
-        dbg_printf("count=%u idle=%u wake_pending=%u", 
-            scoreboard.count, scoreboard.idle,  scoreboard.sb_wake_pending);
         
-        /* Force the manager thread to wakeup if all workers are busy */
-        if (slowpath(scoreboard.idle == 0 && !scoreboard.sb_wake_pending))
-            _wakeup_manager();
+        witem->func(witem->func_arg);    
 
-        // If using RT threads, decrease thread prio if we aren't a high prio queue
-        if (PWQ_RT_THREADS && (queue_priority != WORKQ_HIGH_PRIOQUEUE))
-            ptwq_set_current_thread_priority(queue_priority);
+        atomic_inc(&scoreboard.idle); 
         
-        /* Invoke the callback function, free witem first for possible reuse */
-        func = witem->func;
-        func_arg = witem->func_arg;
         witem_free(witem);
-        
-        func(func_arg);    
-        
-        atomic_inc(&scoreboard.idle); // initial inc was one in worker_start, this is to avoid a race
-
-        // Only take the overhead and change RT priority back if it was not a high priority queue being serviced
-        if (PWQ_RT_THREADS && (queue_priority != WORKQ_HIGH_PRIOQUEUE))
-            ptwq_set_current_thread_priority(WORKQ_HIGH_PRIOQUEUE);
     }
 
     /* NOTREACHED */
@@ -418,7 +476,8 @@ worker_start(void)
 
     dbg_puts("Spawning another worker");
 
-    atomic_inc(&scoreboard.idle);
+    atomic_inc(&pending_thread_create);
+    atomic_inc(&scoreboard.idle); // initialize in idle state to avoid unnecessary thread creation
     atomic_inc(&scoreboard.count);
 
     if (pthread_create(&tid, &detached_attr, worker_main, NULL) != 0) {
@@ -451,10 +510,13 @@ worker_stop(void)
 
         pthread_spin_lock(&workq->mtx);
         
-        do
+        new_mask = atomic_or(&wqlist_mask, wqlist_index_bit);
+        
+        while (slowpath(!(new_mask & wqlist_index_bit)))
         {
-            new_mask = atomic_or(&wqlist_mask, wqlist_index_bit);
-        } while (!(new_mask & wqlist_index_bit));
+            _hardware_pause();
+            new_mask = atomic_or(&wqlist_mask, wqlist_index_bit);                
+        }             
 
         STAILQ_INSERT_TAIL(&workq->item_listhead, witem, item_entry);
 
@@ -476,101 +538,127 @@ worker_stop(void)
 static void *
 manager_main(void *unused __attribute__ ((unused)))
 {
-    unsigned int load_max = cpu_count;
-    unsigned int worker_max, current_thread_count = 0;
+    unsigned int runqueue_length_max = cpu_count; 
+    unsigned int worker_max, threads_total = 0, current_thread_count = 0;
     unsigned int worker_idle_seconds_accumulated = 0;
     unsigned int max_threads_to_stop = 0;
-    unsigned int i;
-    int cond_wait_rv = 0;
+    unsigned int i, idle_surplus_threads = 0;
+    int sem_timedwait_rv = 0;
     sigset_t sigmask;
     struct timespec   ts;
     struct timeval    tp;
 
     worker_max = get_process_limit();
-    scoreboard.load = get_load_average();
+    scoreboard.runqueue_length = get_runqueue_length();
 
     /* Block all signals */
     sigfillset(&sigmask);
     pthread_sigmask(SIG_BLOCK, &sigmask, NULL);
 
     /* Create the minimum number of workers */
-    scoreboard.count = 0;
     for (i = 0; i < worker_min; i++)
         worker_start();
 
     for (;;) {
 
-        pthread_mutex_lock(&scoreboard.sb_wake_mtx);
+        if (scoreboard.sb_suspend == 0) {
+            dbg_puts("manager is sleeping");
 
-        dbg_puts("manager is sleeping");
+            if (gettimeofday(&tp, NULL) != 0) {
+                dbg_perror("gettimeofday()"); // can only fail due to overflow of date > 2038 on 32-bit platforms...
+            }
+
+            /* Convert from timeval to timespec */
+            ts.tv_sec  = tp.tv_sec;
+            ts.tv_nsec = tp.tv_usec * 1000;
+            ts.tv_sec += 1; // wake once per second and check if we have too many idle threads...
+
+            // We should only sleep on the condition if there are no pending signal, spurious wakeup is also ok
+
+            if ((sem_timedwait_rv = sem_timedwait(&scoreboard.sb_sem, &ts)) != 0)
+            {
+                sem_timedwait_rv = errno; // used for ETIMEDOUT below
+                if (errno != ETIMEDOUT)
+                    dbg_perror("sem_timedwait()");
+            }
+
+            dbg_puts("manager is awake");
+        } else {
+            dbg_puts("manager is suspending");
+            if (sem_wait(&scoreboard.sb_sem) != 0)
+                dbg_perror("sem_wait()");
+            dbg_puts("manager is resuming");
+        }
         
-        (void) gettimeofday(&tp, NULL); // TODO - error checking
-        
-        /* Convert from timeval to timespec */
-        ts.tv_sec  = tp.tv_sec;
-        ts.tv_nsec = tp.tv_usec * 1000;
-        ts.tv_sec += 1; // wake once per second and check if we have too many idle threads...
-
-        // We should only sleep on the condition if there are no pending signal, spurious wakeup is also ok
-        if (scoreboard.sb_wake_pending == 0)
-            cond_wait_rv = pthread_cond_timedwait(&scoreboard.sb_wake_cond, &scoreboard.sb_wake_mtx, &ts);
-
-        scoreboard.sb_wake_pending = 0; // we must set this before spawning any new threads below, or we race...
-
-        dbg_puts("manager is awake");
-
-        dbg_printf("load=%u idle=%u workers=%u max_workers=%u worker_min = %u",
-                   scoreboard.load, scoreboard.idle, scoreboard.count, worker_max, worker_min);
+        dbg_printf("idle=%u workers=%u max_workers=%u worker_min = %u",
+                   scoreboard.idle, scoreboard.count, worker_max, worker_min);
                 
         // If no workers available, check if we should create a new one
-        if (scoreboard.idle == 0 && (scoreboard.count > 0)) // last part required for an extremely unlikely race at startup
+        if ((scoreboard.idle == 0) && (scoreboard.count > 0) && (pending_thread_create == 0)) // last part required for an extremely unlikely race at startup
         {
-            scoreboard.load = get_load_average();
-            
-            if ((scoreboard.load < load_max) && (scoreboard.count < worker_max)) 
+            // allow cheap rampup up to worker_idle_threshold without going to /proc / checking run queue length
+            if (scoreboard.count < worker_idle_threshold) 
             {
-                if (scoreboard.count < worker_idle_threshold) // allow cheap rampup up to worker_idle_threshold without going to /proc
+                worker_start();
+            }                
+            else 
+            {
+                // otherwise check if run queue length / stalled threads allows for new creation unless we hit worker_max ceiling
+                
+                if (scoreboard.count < worker_max)
                 {
-                    worker_start();
-                }
-                else // check through /proc, will be a bit more expensive in terms of latency
-                if (threads_runnable(&current_thread_count) == 0)
-                {
-                    // only start thread if we have less runnable threads than cpus
-                    if (current_thread_count >= cpu_count)
+                    if (threads_runnable(&current_thread_count, &threads_total) != 0)
+                        current_thread_count = 0;
+                    
+                    // only start thread if we have less runnable threads than cpus and run queue length allows it
+                    
+                    if (current_thread_count <= cpu_count) // <= discounts the manager thread
                     {
-                        dbg_printf("Not spawning worker thread, thread_runnable = %d >= cpu_count = %d", 
-                                   current_thread_count, cpu_count);
+                        scoreboard.runqueue_length = get_runqueue_length(); 
+
+                        if (scoreboard.runqueue_length <= runqueue_length_max) // <= discounts the manager thread
+                        {
+                            if (scoreboard.idle == 0) // someone might have become idle during getting thread count etc.
+                                worker_start();
+                            else
+                                dbg_puts("skipped thread creation as we got an idle one racing us");
+
+                        }
+                        else
+                        {
+                            dbg_printf("Not spawning worker thread, scoreboard.runqueue_length = %d > runqueue_length_max = %d", 
+                                       scoreboard.runqueue_length, runqueue_length_max);
+                        }
                     }
                     else
                     {
-                        worker_start();
+                        dbg_printf("Not spawning worker thread, thread_runnable = %d > cpu_count = %d", 
+                                   current_thread_count, cpu_count);
                     }
                 }
-                else // always start thread if we can't get runnable count
+                else
                 {
-                    worker_start();
-                }
-            }
-            else // high load, allow rampup up to worker_idle_threshold regardless of this
-            {
-                if (scoreboard.count < worker_idle_threshold) 
-                {
-                    worker_start();
+                    dbg_printf("Not spawning worker thread, scoreboard.count = %d >= worker_max = %d", 
+                               scoreboard.count, worker_max);
                 }                
             }
         }
         else
         {
-            if (cond_wait_rv == ETIMEDOUT) // Only check for ramp down on the 'timer tick'
+            if (sem_timedwait_rv == ETIMEDOUT) // Only check for ramp down on the 'timer tick'
             {
-                if ((scoreboard.idle - worker_idle_threshold) > 0) // only accumulate if there are 'too many' idle threads
+                if (scoreboard.idle > worker_idle_threshold) // only accumulate if there are 'too many' idle threads
                 {
                     worker_idle_seconds_accumulated += scoreboard.idle; // keep track of many idle 'thread seconds' we have
                 
                     dbg_printf("worker_idle_seconds_accumulated = %d, scoreboard.idle = %d, scoreboard.count = %d\n",
                        worker_idle_seconds_accumulated, scoreboard.idle, scoreboard.count);
                 }
+                else
+                {
+                    dbg_puts("Resetting worker_idle_seconds_accumulated");
+                    worker_idle_seconds_accumulated = 0;
+                }
                 
                 // Only consider ramp down if we have accumulated enough thread 'idle seconds'
                 // this logic will ensure that a large number of idle threads will ramp down faster
@@ -579,14 +667,15 @@ manager_main(void *unused __attribute__ ((unused)))
                 if (max_threads_to_stop > 0)
                 {
                     worker_idle_seconds_accumulated = 0; 
-
-                    if (max_threads_to_stop > (scoreboard.idle - worker_idle_threshold))
-                        max_threads_to_stop = (scoreboard.idle - worker_idle_threshold);
+                    idle_surplus_threads = scoreboard.idle - worker_idle_threshold; 
+                    
+                    if (max_threads_to_stop > idle_surplus_threads)
+                        max_threads_to_stop = idle_surplus_threads;
 
                     // Only stop threads if we actually have 'too many' idle ones in the pool
-                    if (scoreboard.idle > worker_idle_threshold)
+                    for (i = 0; i < max_threads_to_stop; i++)
                     {
-                        for (i = 0; i < max_threads_to_stop; i++)
+                        if (scoreboard.idle > worker_idle_threshold)
                         {
                             dbg_puts("Removing one thread from the thread pool");
                             worker_stop();
@@ -594,9 +683,7 @@ manager_main(void *unused __attribute__ ((unused)))
                     }
                 }
             }            
-        }
-        
-        pthread_mutex_unlock(&scoreboard.sb_wake_mtx);
+        }        
     }
 
     /*NOTREACHED*/
@@ -625,12 +712,34 @@ manager_start(void)
     wqlist_has_manager = 1;
 }
 
+void 
+manager_suspend(void)
+{
+    /* Wait for the manager thread to be initialized */
+    while (wqlist_has_manager == 0) {
+        sleep(1);
+    }
+    if (scoreboard.sb_suspend == 0) { 
+        scoreboard.sb_suspend = 1;
+    }
+}
+
+void
+manager_resume(void)
+{
+    if (scoreboard.sb_suspend) { 
+        scoreboard.sb_suspend = 0;
+        __sync_synchronize();
+        (void) sem_post(&scoreboard.sb_sem);        
+    }
+}
+
 void
 manager_workqueue_additem(struct _pthread_workqueue *workq, struct work *witem)
 {
     unsigned int wqlist_index_bit = (0x1 << workq->wqlist_index);
     
-    if (workq->overcommit) {
+    if (slowpath(workq->overcommit)) {
         pthread_t tid;
 
         pthread_mutex_lock(&ocwq_mtx);
@@ -642,6 +751,7 @@ manager_workqueue_additem(struct _pthread_workqueue *workq, struct work *witem)
             dbg_puts("signaling an idle worker");
             pthread_cond_signal(&ocwq_has_work);
             ocwq_idle_threads--;
+            ocwq_signal_count++;
         } else {
             (void)pthread_create(&tid, &detached_attr, overcommit_worker_main, NULL);
         }
@@ -657,10 +767,14 @@ manager_workqueue_additem(struct _pthread_workqueue *workq, struct work *witem)
             // The only possible contention here are with threads performing the same
             // operation on another workqueue, so we will not be blocked long... 
             // Threads operating on the same workqueue will be serialized by the spinlock so it is very unlikely.
-            do
+
+            new_mask = atomic_or(&wqlist_mask, wqlist_index_bit);
+
+            while (slowpath(!(new_mask & wqlist_index_bit)))
             {
-                new_mask = atomic_or(&wqlist_mask, wqlist_index_bit);
-            } while (!(new_mask & wqlist_index_bit));
+                _hardware_pause();
+                new_mask = atomic_or(&wqlist_mask, wqlist_index_bit);                
+            }             
         }
         
         STAILQ_INSERT_TAIL(&workq->item_listhead, witem, item_entry);
@@ -669,11 +783,9 @@ manager_workqueue_additem(struct _pthread_workqueue *workq, struct work *witem)
 
         // Only signal thread wakeup if there are idle threads available
         // and no other thread have managed to race us and empty the wqlist on our behalf already
-        if ((scoreboard.idle > 0)) // && ((wqlist_mask & wqlist_index_bit) != 0)) // disabling this fringe optimization for now
+        if (scoreboard.idle > 0) // && ((wqlist_mask & wqlist_index_bit) != 0)) // disabling this fringe optimization for now
         {
-            pthread_mutex_lock(&wqlist_mtx);
-            pthread_cond_signal(&wqlist_has_work);
-            pthread_mutex_unlock(&wqlist_mtx);
+            pthread_cond_signal(&wqlist_has_work); // don't need to hold the mutex to signal
         }
     }
 }
@@ -687,21 +799,73 @@ get_process_limit(void)
 
     if (getrlimit(RLIMIT_NPROC, &rlim) < 0) {
         dbg_perror("getrlimit(2)");
-        return (50);
+        return (DEFAULT_PROCESS_LIMIT);
     } else {
         return (rlim.rlim_max);
     }
+#elif defined(__sun)
+
+    /* For Solaris we use resource controls as outlined at: */
+    /* http://docs.oracle.com/cd/E19082-01/819-2450/rmctrls.task-3/index.html */
+    /* To enable per-task limits, a project needs to be created and tasks run under this project*/
+    /* The default is essentially unlimited (2147483647), so we clamp it to MIN/MAX for sanity */
+    /* In practice we only support a more strict limit than MAX_PROCESS_LIMIT */
+
+    rctlblk_t *rblk;
+    int num = DEFAULT_PROCESS_LIMIT; // we only use the default if we fail to get it from OS
+    unsigned int threads_total = 0, current_thread_count = 0;
+
+    if ((rblk = (rctlblk_t *)malloc(rctlblk_size())) == NULL) 
+    {
+        dbg_perror("malloc()");
+        return num;
+    }
+    
+    if (getrctl("task.max-lwps", NULL, rblk, RCTL_FIRST) == -1)
+    {
+        dbg_perror("getrctl()");
+    }
+    else
+    {
+        num = rctlblk_get_value(rblk);
+        dbg_printf("task.max-lwps = %u", num);
+    }
+
+    free(rblk);  
+
+    if (threads_runnable(&current_thread_count, &threads_total) == 0)
+        num -= threads_total;  // Discount current number of running threads also
+
+    num -= THREADS_RESERVED;  // when task limits enabled, we will fail to create new threads, so leave some room here
+    
+    if (num < MIN_PROCESS_LIMIT)
+        num = MIN_PROCESS_LIMIT;
+    
+    if (num > MAX_PROCESS_LIMIT)
+        num = MAX_PROCESS_LIMIT;
+            
+    return (unsigned int) (num);
 #else
-    /* Solaris doesn't define this limit anywhere I can see.. */
-    return (64);
+    return (DEFAULT_PROCESS_LIMIT);
 #endif
 }
 
 static unsigned int
-get_load_average(void)
+get_runqueue_length(void)
 {
     double loadavg;
 
+    /* Prefer to use the most recent measurement of the number of running KSEs
+    for Linux and the kstat unix:0:sysinfo: runque/updates ratio for Solaris . */
+
+#if __linux__
+    return linux_get_runqueue_length();
+#elif defined(__sun)
+    return solaris_get_runqueue_length();
+#endif
+
+    /* Fallback to using the 1-minute load average if proper run queue length can't be determined. */
+
     /* TODO: proper error handling */
     if (getloadavg(&loadavg, 1) != 1) {
         dbg_perror("getloadavg(3)");
