@@ -23,11 +23,15 @@
 # include <atomic.h>
 # define atomic_inc      atomic_inc_32
 # define atomic_dec      atomic_dec_32
+# define atomic_inc_nv   atomic_inc_32_nv
+# define atomic_dec_nv   atomic_dec_32_nv
 # define atomic_and      atomic_and_uint_nv
 # define atomic_or      atomic_or_uint_nv
 #else
-# define atomic_inc(p)   __sync_add_and_fetch((p), 1)
-# define atomic_dec(p)   __sync_sub_and_fetch((p), 1)
+# define atomic_inc(p)   (void) __sync_add_and_fetch((p), 1)
+# define atomic_dec(p)   (void) __sync_sub_and_fetch((p), 1)
+# define atomic_inc_nv(p)  __sync_add_and_fetch((p), 1)
+# define atomic_dec_nv(p)  __sync_sub_and_fetch((p), 1)
 # define atomic_and(p,v)   __sync_and_and_fetch((p), (v))
 # define atomic_or(p,v)   __sync_or_and_fetch((p), (v))
 #endif
@@ -39,5 +43,16 @@
 #endif
 #define VISIBLE     __attribute__((visibility("default")))
 
+/*
+ * Android does not provide spinlocks.
+ * See: http://code.google.com/p/android/issues/detail?id=21622
+ */
+#if defined(__ANDROID__)
+#define pthread_spinlock_t     pthread_mutex_t
+#define pthread_spin_lock      pthread_mutex_lock
+#define pthread_spin_unlock    pthread_mutex_unlock
+#define pthread_spin_init(a,b) pthread_mutex_init((a), NULL)
+#define pthread_spin_destroy   pthread_mutex_destroy
+#endif /* defined(__ANDROID__) */
 
 #endif  /* _PTWQ_POSIX_PLATFORM_H */
